//  G. Haase:  Version 2
//  * extended the calculations of Jacobi loop to the calculations before
//  * inialize solution vector u on GPU (set_value_GPU)
//  * get diagonal elements of matrix on GPU (GetDiag_GPU)
//  * Eliminated the memory transfer in scapr_GPU.
//  * Eliminated the memory transfer in the loop
//  * Eliminated all memory transfers for aux. vectors of jacobi iteration
//
//    Scalar product function and kernels derived from Code/CUDA/skalar/skalar_3.cu
//    Remaining kernel functions derived from Code/CUDA/T1/data_mv_xx.cu


#include <iostream>
#include <cmath>
#include <omp.h>                // OpenMP
#ifdef _OPENACC
#include <accel.h>              // OpenACC
#endif
using namespace std;
#include "vdop.h"
#include "getmatrix.h"
#include "jacsolve.h"

//----------------------------------------------------------------------
//    from  CUDA/skalar/skalar_3.cu
//
//    kernel function on device
__global__ void scalar(float *sg, const float *x, const float *y, int N)
{
    extern __shared__ float sdata[];
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    const int str = gridDim.x * blockDim.x;

    float s = 0.0;//x[idx] * y[idx];
    //for (int i = idx + str; i < N; i += str)
    for (int i = idx; i < N; i += str)
        s += x[i] * y[i];
    sdata[threadIdx.x] = s;

    __syncthreads();
    for (int s = blockDim.x >> 1; s > 0; s >>= 1) {
        if (threadIdx.x < s) {
            sdata[threadIdx.x] += sdata[threadIdx.x + s];
        }
        __syncthreads();
    }
    if (threadIdx.x == 0)  sg[blockIdx.x] = sdata[0];
}


//    kernel function for addition on  o n e  block on device
__global__ void add_1_block(float *s, int N)
{
    if (N > blockDim.x) return;
    extern __shared__ float sdata[];
    const int tid = threadIdx.x;
    if (tid < N) sdata[tid] = s[tid];
    else       sdata[tid] = 0.0;

    __syncthreads();
    for (int s = blockDim.x / 2; s > 0; s >>= 1) {
        if (tid < s) {
            sdata[tid] += sdata[tid + s];
        }
        __syncthreads();
    }
    if (tid == 0)  s[0] = sdata[0];
}
//--------------------------------------------------------------------
float dscapr_GPU(const int n, const float x[], const float y[])
{
    cudaDeviceProp prop;
    cudaGetDeviceProperties (&prop, 0);
    const int blocksize = 4 * 64,
//           gridsize = ceil( N / static_cast<float>(blocksize)),
              gridsize = prop.multiProcessorCount,
              resBytes = gridsize * sizeof(float),
              sBytes   = blocksize * sizeof(float) ;
    dim3 dimBlock(blocksize);
    dim3 dimGrid(gridsize);
// allocate memory for temp. vector on device
    float *s_d;               // single number
    cudaMalloc((void **) &s_d, resBytes);


// call the kernel function with  dimGrid.x * dimBlock.x threads
    scalar <<< dimGrid, dimBlock, sBytes>>>(s_d, x, y, n);
//    power of 2 as number of treads per block
    const unsigned int oneBlock = 2 << (int)ceil(log(dimGrid.x + 0.0) / log(2.0));
    //cudaThreadSynchronize();
    add_1_block <<< 1, oneBlock, oneBlock *sizeof(double)>>>(s_d, dimGrid.x);
// copy data:  device --> host
    float sum;
    cudaMemcpy(&sum, s_d, sizeof(double), cudaMemcpyDeviceToHost);

    cudaFree(s_d);
    return sum;
}
// --------------------------------------------------------------------
__global__
void vddiv_GPU(const int n,  float x[], const float y[], const float z[])
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n)
        x[idx] = y[idx] / z[idx];
}

//---------------------------------------------------------------------
__global__
void Defect_GPU(float w[], const float f[], const float u[],
                const int n, const int id[], const int ik[], const float sk[])
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n) {
        float wi = f[idx];
        const int _beg = id[idx], _end = id[idx + 1];
        for (int ij = _beg; ij < _end; ++ij) {
            wi -= sk[ij] * u[ ik[ij] ];
        }
        w[idx] = wi;
    }
}

//---------------------------------------------------------------------
__global__
void vdaxpy_GPU(const int n, float x[], const float y[],
                const float a,   const float z[] )
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n)
        x[idx] = y[idx] + a * z[idx];
}

//---------------------------------------------------------------------
__global__
void set_value_GPU(const int n, float x[], const float a)
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n)
        x[idx] = a;
}
//----------------------------------------------------------------------
__global__
void GetDiag_GPU(const int nnode, const int id[], const int ik[], const float sk[], float d[])
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < nnode)
    {
        //int ip = fetch(idx, idx, id, ik);
        const int id2 = id[idx + 1];
        int ip  = id[idx];
        while (ip < id2 && ik[ip] != idx) {
           ++ip;
        }
        d[idx] = sk[ip];
    }
}
//----------------------------------------------------------------------

void JacobiSolve_GPU(const int nnode,
// 	const int neigh[], const int color,
// 	const MPI::Intracomm& icomm,
                     const int id[], const int ik[], const float sk[],
                     const float f[], float u[] )
{
    //   Data for GPU
    cudaDeviceProp prop;
    cudaGetDeviceProperties (&prop, 0);
    cout << endl << "We work on a " << prop.name << " GPU  with " <<
         prop.multiProcessorCount << " Multiprocessors" << endl;


    const int blocksize = 4 * 64,
              gridsize = ceil( nnode / static_cast<float>(blocksize)),
              //gridsize = prop.multiProcessorCount,
              resBytes = gridsize * sizeof(float),
              sBytes   = blocksize * sizeof(float) ;
    dim3 dimBlock(blocksize);
    dim3 dimGrid(gridsize);

    cout << endl << gridsize << " x " << blocksize << " Threads\n";
    //
    const float omega   = 1.0;
    const int    maxiter = 1000;
    const float tol  = 1e-5,                  /* tolerance   */
                tol2 = tol * tol;             /* tolerance^2 */

    float *dd = new float [nnode];           /* matrix diagonal */
    float *r  = new float [nnode];           /* residual */
    float *w  = new float [nnode];           /* correction */

// *** Data allocate(dd,r,w,u,f,s_d) ***
// allocate memory for vectors on device
    const int nBytes = nnode * sizeof(float);
    float *dd_d, *r_d, *w_d, *u_d, *f_d;  // vectors
    float *s_d;               // single number
    cudaMalloc((void **) &dd_d, nBytes);
    cudaMalloc((void **) &r_d, nBytes);
    cudaMalloc((void **) &w_d, nBytes);
    cudaMalloc((void **) &u_d, nBytes);
    cudaMalloc((void **) &f_d, nBytes);
    cudaMalloc((void **) &s_d, resBytes);
// *** Data allocate(id,ik,sk) ***
// allocate memory for matrix on device
    const int nnz = id[nnode];           // number of non-zero elements
    const int id_Bytes = (nnode + 1) * sizeof(int);
    const int ik_Bytes = nnz * sizeof(int);
    const int sk_Bytes = nnz * sizeof(float);
    int *id_d;
    int *ik_d;
    float *sk_d;
    cudaMalloc((void **) &id_d, id_Bytes);
    cudaMalloc((void **) &ik_d, ik_Bytes);
    cudaMalloc((void **) &sk_d, sk_Bytes);

//   ***  Data  copyin(f,id,ik,sk) ***
// transfer input data:  host --> device
    cudaMemcpy(f_d, f, nBytes, cudaMemcpyHostToDevice);
    cudaMemcpy(id_d, id, id_Bytes, cudaMemcpyHostToDevice);
    cudaMemcpy(ik_d, ik, ik_Bytes, cudaMemcpyHostToDevice);
    cudaMemcpy(sk_d, sk, sk_Bytes, cudaMemcpyHostToDevice);

    {
        cout << endl << " Start Jacobi solver for " << nnode << " d.o.f.s"  << endl;
    }

    GetDiag_GPU<<< dimGrid, dimBlock >>>(nnode, id_d, ik_d, sk_d, dd_d);            /*  diag */

    /*  Choose initial guess  */
    set_value_GPU<<< dimGrid, dimBlock >>>(nnode, u_d, 0.0f);

    /*  Initial sweep  */
    float sigma0, sigma;
    int iter = 0;

    /*  r := f - K*u   */
    Defect_GPU <<< dimGrid, dimBlock >>>(r_d, f_d, u_d, nnode, id_d, ik_d, sk_d);

    /*  w := D^{-1}*r  */
    vddiv_GPU <<< dimGrid, dimBlock >>>(nnode, w_d, r_d, dd_d);/*  w := D^{-1}*r  */

    /* s0 := <w,r>     */
    sigma0 = dscapr_GPU(nnode, w_d, r_d);
    sigma  = sigma0;

    /* Iteration sweeps */
    while ( sigma > tol2 * sigma0 && maxiter > iter) {
        ++iter;
        /*  u := u + om*w  */
        vdaxpy_GPU <<< dimGrid, dimBlock >>>(nnode, u_d, u_d, omega, w_d );

        //*  r := f - K*u   */
        Defect_GPU <<< dimGrid, dimBlock >>>(r_d, f_d, u_d, nnode, id_d, ik_d, sk_d);

        /*  w := D^{-1}*r  */
        vddiv_GPU <<< dimGrid, dimBlock >>>(nnode, w_d, r_d, dd_d);

        /* s := <w,r>     */
        sigma = dscapr_GPU(nnode, w_d, r_d);
    }
//   ***  Data  copyout(u) ***
    // transfer output data from GPU
    cudaMemcpy(u, u_d, nBytes, cudaMemcpyDeviceToHost);

    {
        cout << "aver. Jacobi rate :  " << exp(log(sqrt(sigma / sigma0)) / iter) << "  (" << iter << " iter)" << endl;
        cout << "final error: " << sqrt(sigma / sigma0) << " (rel)   " << sqrt(sigma) << " (abs)\n";
    }
//   ***  Data  deallocate(u) ***
    cudaFree(id_d); cudaFree(ik_d); cudaFree(sk_d);
    cudaFree(w_d); cudaFree(r_d); cudaFree(dd_d);
    cudaFree(f_d); cudaFree(u_d);

    delete [] w;
    delete [] r;
    delete [] dd;
    return;
}

