//		MPI code in C++.
//		See [Gropp/Lusk/Skjellum, "Using MPI", p.33/41 etc.]
//		and  /opt/mpich/include/mpi2c++/comm.h  for details

#include <iostream>
#include <cmath>
#include <ctime>
//#include <omp.h>                // OpenMP
using namespace std;
#include "geom.h"
#include "getmatrix.h"
#include "jacsolve_gpu.h"
#include "userset.h"

int main(int , char** )
{
    const int numprocs = 1;
    const int myrank   = 0;

    if (myrank == 0)
    { cout << "\n There are " << numprocs << " MPI processes running.\n \n"; }

    const int procx = static_cast<int>(sqrt(numprocs+0.0));
    const int procy = procx;

    if (procy * procx != numprocs)
    {
      cout << "\n Wrong number of processors !\n \n";
    }
    else
    {
      //int nthreads =  status_threads();

      int neigh[4], color;
      IniGeom(myrank, procx, procy, neigh, color); // Topology of domain decomposition

      float xl,xr,yb,yt;                          // subdomain coordinates
      IniCoord(myrank, procx, procy, xl, xr, yb, yt);

      //bool ScaleUp = !true;
      int nx,ny,NXglob,NYglob;  /* number of local intervals on (xl,xr)=:nx, (yb,yt)=:ny */
	nx = 100;
	ny = 100;
	NXglob = nx*procx;
	NYglob = ny*procy;
    cout << "Intervalls: " << NXglob << " x " << NYglob << endl;

    // #########################################################################
// Initialze the mesh
      int nnode, nelem;        // number of ccordinates, number of triangular elements
      float *xc;              // coordinates         xc[nnode,2]
      int    *ia;              // element conectivity ia[nelem,3]
      GetMesh(nx, ny, xl, xr, yb, yt, nnode, xc, nelem, ia);
      //DebugMesh(nnode, xc, nelem, 3, ia);  {int ijk; cin >> ijk;}

// Matrix Graph
      int nnz(-1);                           // number of non-zero eelemnts in matrix
      float *sk;                            // sk[nnz]  non-zero entries
      int    *ik,                            // ik[nnz]  column indices
             *id;                            // id[nnode+1] index of row begin
      Get_Matrix_Pattern(nelem, 3, ia, nnz, id, ik, sk);
      //DebugMatrix(nnode, id, ik, sk);  {int ijk; cin >> ijk;}

      /* allocate memory */
      float *u = new float [nnode],        /* temperature       */
            *f = new float [nnode];        /* r.h.s.            */

      /* matrix, rhs generation including Dirichlet B.C. */
        //double ts = omp_get_wtime();
        double ts = clock();
      GetMatrix (nelem, 3, ia, nnode, xc, nnz, id, ik, sk, f);
        //double ts1 = omp_get_wtime() - ts;                 // OpenMP timing
        double ts1 = (clock() - ts)/CLOCKS_PER_SEC;                 // timing
        cout << "\nGetMatrix: OpenMP timing in sec. : " << ts1 << endl;
      //DebugMatrix(nnode, id, ik, sk);  {int ijk; cin >> ijk;}


      SetU(nx, ny, u);                       /* set solution/boundary values */
      ApplyDirichletBC(nx, ny, neigh, u, id, ik, sk, f); /* apply B.C.       */

// *********************************************************************
//    Transfer/allocate data outside function JacobiSolve_GPU
// *********************************************************************
    //   Data for GPU
    cudaDeviceProp prop;
    cudaGetDeviceProperties (&prop, 0);
    cout << endl << "We work on a " << prop.name << " GPU  with " <<
         prop.multiProcessorCount << " Multiprocessors" << endl;

    const int blocksize = 4 * 64,
              gridsize = ceil( nnode / static_cast<float>(blocksize));
              //gridsize = prop.multiProcessorCount,
              //resBytes = gridsize * sizeof(float),
              //sBytes   = blocksize * sizeof(float) ;
    dim3 dimBlock(blocksize);
    dim3 dimGrid(gridsize);

    cout << endl << gridsize << " x " << blocksize << " Threads\n";

// *** Data allocate(dd,r,w,u,f,s_d) ***
// allocate memory for vectors on device
    const int nBytes = nnode * sizeof(float);
    float *dd_d, *r_d, *w_d, *u_d, *f_d;  // vectors
    cudaMalloc((void **) &dd_d, nBytes);           /* matrix diagonal */
    cudaMalloc((void **) &r_d, nBytes);            /* residual */
    cudaMalloc((void **) &w_d, nBytes);            /* correction */
    cudaMalloc((void **) &u_d, nBytes);            /* solution */
    cudaMalloc((void **) &f_d, nBytes);            /* rhs */
// *** Data allocate(id,ik,sk) ***
// allocate memory for matrix on device
    //const int nnz = id[nnode];           // number of non-zero elements
    const int id_Bytes = (nnode + 1) * sizeof(int);
    const int ik_Bytes = nnz * sizeof(int);
    const int sk_Bytes = nnz * sizeof(float);
    int *id_d;
    int *ik_d;
    float *sk_d;
    cudaMalloc((void **) &id_d, id_Bytes);
    cudaMalloc((void **) &ik_d, ik_Bytes);
    cudaMalloc((void **) &sk_d, sk_Bytes);

//   ***  Data  copyin(f,id,ik,sk) ***
// transfer input data:  host --> device
    cudaMemcpy(f_d, f, nBytes, cudaMemcpyHostToDevice);
    cudaMemcpy(id_d, id, id_Bytes, cudaMemcpyHostToDevice);
    cudaMemcpy(ik_d, ik, ik_Bytes, cudaMemcpyHostToDevice);
    cudaMemcpy(sk_d, sk, sk_Bytes, cudaMemcpyHostToDevice);

// *********************************************************************
//     Call solver
// *********************************************************************

      //double tstart = omp_get_wtime();                       // OpenMP timing
      double tstart = clock();                       // timing

      JacobiSolve_GPU_4(nnode, id_d, ik_d, sk_d, f_d, u_d,
                        r_d, w_d, dd_d, dimGrid, dimBlock );

      //double t1 = omp_get_wtime() - tstart;                 // OpenMP timing
      double t1 = (clock() - tstart)/CLOCKS_PER_SEC;                 // timing
      cout << "JacobiSolve: OpenMP timing in sec. : " << t1 << endl;

// *********************************************************************
//   Transfer solution from device
// *********************************************************************
//   ***  Data  copyout(u) ***
    // transfer output data from GPU
    cudaMemcpy(u, u_d, nBytes, cudaMemcpyDeviceToHost);
//   ***  Data  deallocate(u) ***
    cudaFree(id_d); cudaFree(ik_d); cudaFree(sk_d);
    cudaFree(w_d); cudaFree(r_d); cudaFree(dd_d);
    cudaFree(f_d); cudaFree(u_d);
// *********************************************************************

      //SaveVectorP(myrank, "t.dat", u, nx, ny, xl, xr, yb, yt); // solution for gnuplot

      delete [] ia; delete [] xc;
      delete [] id; delete [] ik; delete [] sk; delete [] f; delete [] u;
  }
  return 0;
}



