//  G. Haase: Version -- originates from  CUDA/jacobi
//  All data transfer/allocation are done outside function JacobiSolve_GPU
//  in the main programm and before timing.

#include <iostream>
#include <cmath>
#include <mpi.h>                // MPI
using namespace std;
//#include "vdop.h"
#include "getmatrix.h"
#include "vecaccu.h"

//----------------------------------------------------------------------
//    from  CUDA/skalar/skalar_3.cu
//
//    kernel function on device
__global__ void scalar(float *sg, const float *x, const float *y, int N)
{
    extern __shared__ float sdata[];
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    const int str = gridDim.x * blockDim.x;

    float s = 0.0;
    for (int i = idx; i < N; i += str)
        s += x[i] * y[i];
    sdata[threadIdx.x] = s;

    __syncthreads();
    for (int s = blockDim.x >> 1; s > 0; s >>= 1) {
        if (threadIdx.x < s) {
            sdata[threadIdx.x] += sdata[threadIdx.x + s];
        }
        __syncthreads();
    }
    if (threadIdx.x == 0)  sg[blockIdx.x] = sdata[0];
}


//    kernel function for addition on  o n e  block on device
__global__ void add_1_block(float *s, int N)
{
    if (N > blockDim.x) return;
    extern __shared__ float sdata[];
    const int tid = threadIdx.x;
    if (tid < N) sdata[tid] = s[tid];
    else       sdata[tid] = 0.0;

    __syncthreads();
    for (int s = blockDim.x / 2; s > 0; s >>= 1) {
        if (tid < s) {
            sdata[tid] += sdata[tid + s];
        }
        __syncthreads();
    }
    if (tid == 0)  s[0] = sdata[0];
}
//--------------------------------------------------------------------
float dscapr_GPU(const int n, const float x[], const float y[])
{
    cudaDeviceProp prop;
    cudaGetDeviceProperties (&prop, 0);
    const int blocksize = 4 * 64,
//           gridsize = ceil( N / static_cast<float>(blocksize)),
              gridsize = prop.multiProcessorCount,
              resBytes = gridsize * sizeof(float),
              sBytes   = blocksize * sizeof(float) ;
    dim3 dimBlock(blocksize);
    dim3 dimGrid(gridsize);
// allocate memory for temp. vector on device
    float *s_d;               // single number
    cudaMalloc((void **) &s_d, resBytes);


// call the kernel function with  dimGrid.x * dimBlock.x threads
    scalar <<< dimGrid, dimBlock, sBytes>>>(s_d, x, y, n);
//    power of 2 as number of treads per block
    const unsigned int oneBlock = 2 << (int)ceil(log(dimGrid.x + 0.0) / log(2.0));
    //cudaThreadSynchronize();
    add_1_block <<< 1, oneBlock, oneBlock *sizeof(float)>>>(s_d, dimGrid.x);
// copy data:  device --> host
    float sum;
    cudaMemcpy(&sum, s_d, sizeof(float), cudaMemcpyDeviceToHost);

    cudaFree(s_d);
    return sum;
}
// --------------------------------------------------------------------
__global__
void vddiv_GPU(const int n,  float x[], const float y[], const float z[])
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n)
        x[idx] = y[idx] / z[idx];
}

//---------------------------------------------------------------------
__global__
void Defect_GPU(float w[], const float f[], const float u[],
                const int n, const int id[], const int ik[], const float sk[])
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n) {
        float wi = f[idx];
        const int _beg = id[idx], _end = id[idx + 1];
        for (int ij = _beg; ij < _end; ++ij) {
            wi -= sk[ij] * u[ ik[ij] ];
        }
        w[idx] = wi;
    }
}

//---------------------------------------------------------------------
__global__
void vdaxpy_GPU(const int n, float x[], const float y[],
                const float a,   const float z[] )
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n)
        x[idx] = y[idx] + a * z[idx];
}

//---------------------------------------------------------------------
__global__
void set_value_GPU(const int n, float x[], const float a)
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < n)
        x[idx] = a;
}
//----------------------------------------------------------------------
__global__
void GetDiag_GPU(const int nnode, const int id[], const int ik[], const float sk[], float d[])
{
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < nnode)
    {
        //int ip = fetch(idx, idx, id, ik);
        const int id2 = id[idx + 1];
        int ip  = id[idx];
        while (ip < id2 && ik[ip] != idx) {
           ++ip;
        }
        d[idx] = sk[ip];
        //if (ip>=id2) d[idx] = -9876;
    }
}

//----------------------------------------------------------------------

float skalar(const int n, const float x[], const float y[], const MPI::Intracomm& icomm)
{
  //const float s = dscapr(n,x,y);                // call sequential inner product
  const float s = dscapr_GPU(n, x, y);            // call GPU inner product
        float sg;
  icomm.Allreduce(&s,&sg,1,MPI::FLOAT,MPI::SUM);

  return(sg);
}

//----------------------------------------------------------------------

void JacobiSolve_GPU_4(const int nnode, const int nx, const int ny,
	  const int neigh[], const int color, const MPI::Intracomm& icomm,
      const int id_d[], const int ik_d[], const float sk_d[],
      const float f_d[], float u_d[],
      float r_d[], float w_d[], float dd_d[],
      const dim3 dimGrid, const dim3 dimBlock )
{
    float *tmp = new float [10*nnode];         // aux. array for storing GPU-array on host
    const int nBytes = nnode*sizeof(tmp[0]);   // bytes to transfer between GPU/host

    const float omega   = 1.0;
    const int    maxiter = 1000;
    const float tol  = 1e-5,                  /* tolerance   */
                tol2 = tol * tol;             /* tolerance^2 */

    if (icomm.Get_rank()==0)
    {
      cout << endl << " Start Jacobi solver for " << nnode << " d.o.f.s"  << endl;
    };
    fflush(stdout);icomm.Barrier();

    GetDiag_GPU<<< dimGrid, dimBlock >>>(nnode, id_d, ik_d, sk_d, dd_d);            /*  diag */
    //cudaMemcpy(tmp, dd_d, nBytes, cudaMemcpyDeviceToHost);
    //DebugVector(nnode, tmp);  {int ijk; cin >> ijk;}
            //cout << "GetDiag  " <<  dscapr_GPU(nnode, dd_d, dd_d) << endl;
//  accu dd
    cudaMemcpy(tmp, dd_d, nBytes, cudaMemcpyDeviceToHost);
    VecAccu(nx, ny, tmp, neigh, color, icomm);
    cudaMemcpy(dd_d, tmp, nBytes, cudaMemcpyHostToDevice);



    /*  Choose initial guess  */
    set_value_GPU<<< dimGrid, dimBlock >>>(nnode, u_d, 0.0f);

    /*  Initial sweep  */
    float sigma0, sigma;
    int iter = 0;

    /*  r := f - K*u   */
    Defect_GPU <<< dimGrid, dimBlock >>>(r_d, f_d, u_d, nnode, id_d, ik_d, sk_d);

    /*  w := D^{-1}*r  */
    vddiv_GPU <<< dimGrid, dimBlock >>>(nnode, w_d, r_d, dd_d);/*  w := D^{-1}*r  */
//  accu w
    cudaMemcpy(tmp, w_d, nBytes, cudaMemcpyDeviceToHost);
    VecAccu(nx, ny, tmp, neigh, color, icomm);
    cudaMemcpy(w_d, tmp, nBytes, cudaMemcpyHostToDevice);

    /* s0 := <w,r>     */
    //sigma0 = dscapr_GPU(nnode, w_d, r_d);
    sigma0 = skalar(nnode, w_d, r_d,icomm);                 // accumulates internally
    sigma  = sigma0;

    /* Iteration sweeps */
    while ( sigma > tol2 * sigma0 && maxiter > iter) {
        ++iter;
        /*  u := u + om*w  */
        vdaxpy_GPU <<< dimGrid, dimBlock >>>(nnode, u_d, u_d, omega, w_d );

        //*  r := f - K*u   */
        Defect_GPU <<< dimGrid, dimBlock >>>(r_d, f_d, u_d, nnode, id_d, ik_d, sk_d);

        /*  w := D^{-1}*r  */
        vddiv_GPU <<< dimGrid, dimBlock >>>(nnode, w_d, r_d, dd_d);
//  accu w
        cudaMemcpy(tmp, w_d, nBytes, cudaMemcpyDeviceToHost);
        VecAccu(nx, ny, tmp, neigh, color, icomm);
        cudaMemcpy(w_d, tmp, nBytes, cudaMemcpyHostToDevice);        /* s := <w,r>     */

        //sigma = dscapr_GPU(nnode, w_d, r_d);
        sigma = skalar(nnode, w_d, r_d,icomm);              // accumulates internally
// accu sigma
    }

    fflush(stdout);icomm.Barrier();

    if (icomm.Get_rank()==0)
    {
        cout << "aver. Jacobi rate :  " << exp(log(sqrt(sigma / sigma0)) / iter) << "  (" << iter << " iter)" << endl;
        cout << "final error: " << sqrt(sigma / sigma0) << " (rel)   " << sqrt(sigma) << " (abs)\n";
    }
    fflush(stdout);icomm.Barrier();



    return;
}

