#ifndef JACSOLVEGPU_FILE
#define JACSOLVEGPU_FILE
#include <mpi.h>

/**
 * Solves linear system of equations  K @p u = @p f  via the Jacobi iteration.
 * We use a distributed symmetric  CSR matrix K(@p sk,@p id, @p ik) and initial guess of the
 * solution is set to 0.
 * @param[in] nnode	number of nodes and number of rows
 * @param[in] nx	local number intervals in x-direction
 * @param[in] ny	local number intervals in x-direction
 * @param[in] neigh	neighbors in parallel
 * @param[in] color	my color in parallel
 * @param[in] sk_d	on GPU: vector non-zero entries of CSR matrix
 * @param[in] id_d	on GPU: index vector containing the last entry in a CSR row
 * @param[in] ik_d	on GPU: column index vector of CSR matrix
 * @param[in] f_d	on GPU: distributed local vector storing the right hand side
 * @param[out] u_d	on GPU: accumulated local vector storing the solution.
 * @param[inout] r_d  on GPU: residuum (aux. vector)
 * @param[inout] w_d  on GPU: correction (aux. vector)
 * @param[inout] dd_d on GPU: diagonale values(aux. vector)
 * @param[in] dimGrid  number of blockthreads
 * @param[in] dimBlock number of threads per block
*/

void JacobiSolve_GPU_4(const int nnode, const int nx, const int ny,
	 const int neigh[], const int color, const MPI::Intracomm& icomm,
     const int id_d[], const int ik_d[], const float sk_d[],
     const float f_d[], float u_d[],
     float r_d[], float w_d[], float dd_d[],
     const dim3 dimGrid, const dim3 dimBlock );
/**
 * Solves linear system of equations  K @p u = @p f  via the Jacobi iteration.
 * We use a distributed symmetric  CSR matrix K(@p sk,@p id, @p ik) and initial guess of the
 * solution is set to 0.
 * @param[in] nnode	number of nodes and number of rows
 * @param[in] sk	vector non-zero entries of CSR matrix
 * @param[in] id	index vector containing the last entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] f		distributed local vector storing the right hand side
 * @param[out] u	accumulated local vector storing the solution.
*/

void JacobiSolve_GPU(const int nnode, const int id[], const int ik[], const float sk[],
                 const float f[], float u[] );

/** @brief  Calculates the Euclidian inner product of two vectors on GPU.
 *
 * @param[in]  n  number of elements in both vectors.
 * @param[in]  x source vector on GPU
 * @param[in]  y source vector on GPU
 * @return Euclidian inner product @f$\langle x,y \rangle@f$ on GPU
 *
 */
float dscapr_GPU(const int n, const float x[], const float y[]);


/** @brief  Element-wise vector divison x = y/z.
 *
 * @param[in]  n  number of elements to divide
 * @param[out] x  target vector
 * @param[in]  y  source vector
 * @param[in]  z  source vector
 *
 */
void vddiv_GPU(const int n,  float x[], const float y[], const float z[]);


/** @brief  Element-wise vector multiplication x = y*z.
 *  A stride @p iy =0 for the source stride adds a constant value stored in @p y[0].
 *  The same holds for vector @p z.
 *
 * @param[in]  n  number of elements to multiply
 * @param[out] x  target vector
 * @param[in]  ix stride for target vector @p x
 * @param[in]  y  source vector
 * @param[in]  iy stride for source vector @p y
 * @param[in]  z  source vector
 * @param[in]  iz stride for source vector @p z
 *
 */

void vdaxpy_GPU(const int n, float x[], const float y[],
                   const float a,   const float z[] );


void set_value_GPU(const int n, float x[], const float a);


void Defect_GPU(float w[], const float f[], const float u[],
            const int nnode, const int id[], const int ik[], const float sk[]);


/**
 * Copies the diagonal entries of a CSR matrix (@p sk,@p id, @p ik)
 * onto a vector @p d.
 * @param[in] nnode	number of rows
 * @param[in] id	index vector containing the last entry in a CSR row
 * @param[in] ik	column index vector of CSR matrix
 * @param[in] sk	vector non-zero entries of CSR matrix
 * @param[out] d	accumulated diagonal entries of CSR matrix
*/
void GetDiag_GPU(const int nnode, const int id[], const int ik[], const float sk[], float d[]);


#endif

