// scalar product: 2: recursive local summation
#include <cassert>
#include <iostream>
#include <ctime>
using namespace std;
#include <cutil_inline.h>

//    kernel function on device  
__global__ void scalar(double *sg, const double *x, const double *y, int N)
{
  extern __shared__ double sdata[];
  int idx = blockIdx.x * blockDim.x + threadIdx.x;
  int str = gridDim.x*blockDim.x;
  
  double s = x[idx]*y[idx];
  for (int i=idx+str; i<N; i+=str)
      s+= x[i]*y[i];
  sdata[threadIdx.x] = s;
  
  __syncthreads();
  for (unsigned int s=blockDim.x/2; s>0; s>>=1) {
     if (threadIdx.x < s) {
         sdata[threadIdx.x] += sdata[threadIdx.x + s];
      }
  __syncthreads();
  }
  if (threadIdx.x == 0)  sg[blockIdx.x] = sdata[0];
}


//  main function on host
int main(void)
{
 double *x_h, *y_h, *s_h; // host data
 double *x_d, *y_d, *s_d; // device data
 double sum, tstart, tgpu;
 const int N = 14000000,
           nBytes = N*sizeof(double);
 const int LOOP = 100;
 
 
 cudaDeviceProp prop;
 cudaGetDeviceProperties (&prop, 0);
 cout<< endl << "We work on a "<< prop.name << " GPU  with " << 
       prop.multiProcessorCount << " Multiprocessors"<< endl;
 
           
 const int blocksize= 4*64, 
//           gridsize = ceil( N / static_cast<double>(blocksize)),
           gridsize = prop.multiProcessorCount,
           resBytes = gridsize*sizeof(double),
           sBytes   = blocksize*sizeof(double) ; 
 dim3 dimBlock(blocksize);
 dim3 dimGrid(gridsize);           

 cout << endl << gridsize << " x " << blocksize << " Threads\n";

 x_h = new double [N];
 y_h = new double [N];
 s_h = new double [gridsize];
// allocate memory on device
 cudaMalloc((void **) &x_d, nBytes);
 cudaMalloc((void **) &y_d, nBytes);
 cudaMalloc((void **) &s_d, resBytes);

 double ss=0.0;
 for (int i=0; i<N; i++) 
   { x_h[i] = (i % 137)+1; y_h[i] = 1.0/x_h[i]; ss+=x_h[i]*y_h[i];}

// copy data:  host --> device
 cutilSafeCall(cudaMemcpy(x_d, x_h, nBytes, cudaMemcpyHostToDevice));
 cudaMemcpy(y_d, y_h, nBytes, cudaMemcpyHostToDevice);

 tstart = clock();
 for (int k=0; k<LOOP; ++k) {
// call the kernel function with  dimGrid.x * dimBlock.x threads 
    scalar<<<dimGrid, dimBlock, sBytes>>>(s_d, x_d, y_d, N);
// copy data:  device --> host
    cutilSafeCall(cudaMemcpy(s_h, s_d, resBytes, cudaMemcpyDeviceToHost));
 
    sum=0.0;
    for (int i=0; i<gridsize; i++) sum+=s_h[i];
    
}
 tgpu = (clock() - tstart)/CLOCKS_PER_SEC/LOOP;
 cout << "GPU: " << tgpu << " sec.; " << 
   2e-9*N/tgpu << " GFLOPS;  " << 
   2.0*N*sizeof(double)/tgpu/1024/1024/1024 << " GiB/sec.\n ";

 if (static_cast<int>(sum)!=N)
    {
    cout << " !!  W R O N G  result  " << endl;
    cout << sum << "  vs.  " << N << "(" << ss<< ")" << endl;
    }
 else
    cout << " correct  result  " << endl;

 delete [] x_h; delete [] y_h; delete [] s_h; 
 cudaFree(x_d); cudaFree(y_d); cudaFree(s_d);
 return 0;
}
