//		MPI code in C++.
//		See [Gropp/Lusk/Skjellum, "Using MPI", p.33/41 etc.]
//		and  /opt/mpich/include/mpi2c++/comm.h  for details

#include <iostream>
#include <cmath>
#include <omp.h>                // OpenMP
#ifdef _OPENACC
#include <accel.h>              // OpenACC
#endif
using namespace std;
#include "geom.h"
#include "getmatrix.h"
#include "jacsolve.h"
#include "userset.h"
#include "help.h"

int main(int , char** )
{
    const int numprocs = 1;
    const int myrank   = 0;

    if (myrank == 0)
    { cout << "\n There are " << numprocs << " MPI processes running.\n \n"; }

    const int procx = static_cast<int>(sqrt(numprocs+0.0));
    const int procy = procx;

    if (procy * procx != numprocs)
    {
      cout << "\n Wrong number of processors !\n \n";
    }
    else
    {
      int nthreads =  status_threads();

      int neigh[4], color;
      IniGeom(myrank, procx, procy, neigh, color); // Topology of domain decomposition

      float xl,xr,yb,yt;                          // subdomain coordinates
      IniCoord(myrank, procx, procy, xl, xr, yb, yt);

      //bool ScaleUp = !true;
      int nx,ny,NXglob,NYglob;  /* number of local intervals on (xl,xr)=:nx, (yb,yt)=:ny */
	nx = 100;
	ny = 100;
	NXglob = nx*procx;
	NYglob = ny*procy;
    cout << "Intervalls: " << NXglob << " x " << NYglob << endl;

    // #########################################################################
#ifdef _OPENACC
    acc_init(acc_device_nvidia);             // OpenACC call
    cout <<"\nOpenACC Init done.\n";
#endif

// Initialze the mesh
      int nnode, nelem;        // number of ccordinates, number of triangular elements
      float *xc;              // coordinates         xc[nnode,2]
      int    *ia;              // element conectivity ia[nelem,3]
      GetMesh(nx, ny, xl, xr, yb, yt, nnode, xc, nelem, ia);
      //DebugMesh(nnode, xc, nelem, 3, ia);  {int ijk; cin >> ijk;}

// Matrix Graph
      int nnz(-1);                           // number of non-zero eelemnts in matrix
      float *sk;                            // sk[nnz]  non-zero entries
      int    *ik,                            // ik[nnz]  column indices
             *id;                            // id[nnode+1] index of row begin
      Get_Matrix_Pattern(nelem, 3, ia, nnz, id, ik, sk);
      //DebugMatrix(nnode, id, ik, sk);  {int ijk; cin >> ijk;}

      /* allocate memory */
      float *u = new float [nnode],        /* temperature       */
            *f = new float [nnode];        /* r.h.s.            */

      //float tstart = omp_get_wtime();                       // OpenMP timing

      /* matrix, rhs generation including Dirichlet B.C. */
        double ts = omp_get_wtime();
      GetMatrix (nelem, 3, ia, nnode, xc, nnz, id, ik, sk, f);
        double ts1 = omp_get_wtime() - ts;                 // OpenMP timing
        cout << "\nGetMatrix: OpenMP timing in sec. : " << ts1 << endl;
      //DebugMatrix(nnode, id, ik, sk);  {int ijk; cin >> ijk;}


      SetU(nx, ny, u);                       /* set solution/boundary values */
      ApplyDirichletBC(nx, ny, neigh, u, id, ik, sk, f); /* apply B.C.       */

      double tstart = omp_get_wtime();                       // OpenMP timing

      JacobiSolve(nnode, id, ik, sk, f, u );

      double t1 = omp_get_wtime() - tstart;                 // OpenMP timing
      cout << "JacobiSolve: OpenMP timing in sec. : " << t1 << endl;


      SaveVectorP(myrank, "t.dat", u, nx, ny, xl, xr, yb, yt); // solution for gnuplot

      delete [] ia; delete [] xc;
      delete [] id; delete [] ik; delete [] sk; delete [] f; delete [] u;
  }
  return 0;
}



