#include <iostream>
#include <cmath>
#include <omp.h>                // OpenMP
#ifdef _OPENACC
#include <accel.h>              // OpenACC
#endif
using namespace std;
#include "vdop.h"
#include "debugd.h"
#include "getmatrix.h"
#include "jacsolve.h"
#include "accudiag.h"
#include "vecaccu.h"
#include "skalar.h"

void JacobiSolve(const int nnode,const int nx, const int ny,
 	const int neigh[], const int color, const MPI_Comm icomm,
                 const int id[], const int ik[], const float sk[],
                 const float f[], float u[] )
{
    int myrank;
    MPI_Comm_rank(icomm,&myrank);

    const float omega   = 1.0;
    const int    maxiter = 1000;
    //const int    maxiter = 1;
    const float tol  = 1e-5,                  /* tolerance   */
                 tol2 = tol * tol;             /* tolerance^2 */

    float *dd = new float [nnode],           /* matrix diagonal */
           *r  = new float [nnode],           /* residual */
           *w  = new float [nnode];           /* correction */

    #pragma acc data copy(u[0:nnode]) copyin(f[0:nnode],id[0:nnode+1],ik[0:id[nnode]],sk[0:id[nnode]]) \
     create(dd[0:nnode],r[0:nnode],w[0:nnode])
  {
        if (myrank==0)
        {
            cout << endl << " Start Jacobi solver for " << nnode << " d.o.f.s"  << endl;
        }
                  // acc update in VeccAccu
        AccuDiag(nnode, id, ik, sk, dd, nx, ny, neigh, color, icomm); /* Acuum. diag */
        //DebugD(nnode, dd);

        float sigma0,sigma;
        int iter;
        /*  Choose initial guess  */
        #pragma omp parallel for
        #pragma acc kernels pcopy(u[0:nnode])
        for (int k=0; k<nnode; ++k) { u[k] = 0.0; }  /*  u := 0         */
                                                   /*  Initial sweep  */
        Defect(r, f, u, nnode, id, ik, sk);        /*  r := f - K*u   */

        #pragma omp parallel for
        #pragma acc kernels loop independent pcopyout(w[0:nnode]) pcopyin(r[0:nnode], dd[0:nnode])
        for (int k=0; k<nnode; ++k) { w[k] = r[k]/dd[k]; }  /*  w := D^{-1}*r  */
        VecAccu(nx, ny, w, neigh, color, icomm);	 /*  dist.-->acc.   */

        iter  = 0;
        //sigma0 = dscapr(nnode, w, r);              /* s0 := <w,r>     */
        sigma0 = skalar(nnode, w, r, icomm);	 /* s0 := <w,r>     */
        sigma  = sigma0;


        /* Iteration sweeps */
        while ( sigma > tol2 * sigma0 && maxiter > iter) {
            ++iter;
            vdaxpy(nnode, u, u, omega, w );          /*  u := u + om*w  */
            Defect(r, f, u, nnode, id, ik, sk);      /*  r := f - K*u   */

            #pragma omp parallel for
            #pragma acc kernels loop independent pcopyout(w[0:nnode]) pcopyin(r[0:nnode], dd[0:nnode])
            for (int k=0; k<nnode; ++k) { w[k] = r[k]/dd[k]; }  /*  w := D^{-1}*r  */
                    // acc update in VeccAccu
            VecAccu(nx, ny, w, neigh, color, icomm);	 /*  dist.-->acc.   */

            sigma = skalar(nnode, w, r, icomm);		     /* s := <w,r>     */
            //if ( omp_get_thread_num()==0) cout << "Iteration " << iter << " : " << sqrt(sigma/sigma0) << endl;
            #pragma omp barrier
        }
        {
           MPI_Barrier(icomm);fflush(stdout);MPI_Barrier(icomm);
           if (myrank==0)
           {
              cout << "aver. Jacobi rate :  " << exp(log(sqrt(sigma/sigma0))/iter) << "  (" << iter << " iter)"<< endl;
              cout << "final error: " << sqrt(sigma / sigma0) << " (rel)   " << sqrt(sigma) << " (abs)\n";
           }
           MPI_Barrier(icomm);fflush(stdout);MPI_Barrier(icomm);
        }
    } // end: acc data


    delete [] w;
    delete [] r;
    delete [] dd;
    return;
}

